/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// vid_glx.c -- OpenGL Extension X Window System video driver

#define GLX_GLXEXT_PROTOTYPES
#include <GL/glx.h>

#include "quakedef.h"
#include "xquake.h"
#include "quake.xpm"

#define WARP_WIDTH 320
#define WARP_HEIGHT 200

static int scrnum;
static GLXContext ctx = NULL;

Atom wm_delete_window_atom;

viddef_t vid; // global video state

cvar_t vid_mode = {"vid_mode","0",false};

static XF86VidModeModeInfo init_vidmode, game_vidmode;

//==========================================================================
//
//  HARDWARE GAMMA
//
//==========================================================================

unsigned short	 vid_gammaramp[3][256];
unsigned short	 vid_systemgammaramp[3][256]; // to restore gamma
qboolean vid_gammaworks;

/*
================
VID_Gamma_Set -- apply gamma correction
================
*/
void VID_Gamma_Set (void)
{
	if (!vid_gammaworks)
		return;

	if (!XF86VidModeSetGammaRamp (x_disp, scrnum, 256, vid_gammaramp[0], vid_gammaramp[1], vid_gammaramp[2]))
		Con_Printf ("VID_Gamma_Set: failed on XF86VidModeSetGammaRamp\n");

//	Con_Printf ("VID_Gamma_Set %g\n", v_gamma.value);
}

/*
================
VID_Gamma_Restore -- restore system gamma
================
*/
void VID_Gamma_Restore (void)
{
	if (!vid_gammaworks)
		return;

	if (!XF86VidModeSetGammaRamp (x_disp, scrnum, 256, vid_systemgammaramp[0], vid_systemgammaramp[1], vid_systemgammaramp[2]))
		Con_Printf ("VID_Gamma_Restore: failed on XF86VidModeSetGammaRamp\n");

//	Con_Printf ("VID_Gamma_Restore\n");
}

/*
================
VID_Gamma_Shutdown -- called on exit
================
*/
void VID_Gamma_Shutdown (void)
{
	VID_Gamma_Restore ();
}

/*
================
VID_Gamma -- call when the cvar changes
================
*/
void VID_Gamma (void)
{
	int i;

	for (i=0; i<256; i++)
		vid_gammaramp[0][i] = vid_gammaramp[1][i] = vid_gammaramp[2][i] =
			CLAMP(0, (int) (255 * pow ((i+0.5)/255.5, v_gamma.value) + 0.5), 255) << 8;

	VID_Gamma_Set ();
}

/*
================
VID_Gamma_Init -- call on init
================
*/
void VID_Gamma_Init (void)
{
	int	xf86vm_gammaramp_size;

	vid_gammaworks = false;

	XF86VidModeGetGammaRampSize (x_disp, scrnum, &xf86vm_gammaramp_size);
	vid_gammaworks = (xf86vm_gammaramp_size == 256);
	if (vid_gammaworks)
		XF86VidModeGetGammaRamp (x_disp, scrnum, xf86vm_gammaramp_size, vid_systemgammaramp[0], vid_systemgammaramp[1], vid_systemgammaramp[2]);

	if (!vid_gammaworks)
		Con_Printf ("Hardware gamma unavailable\n");
}

//
// direct draw software compatability stuff
//
/*
================
D_BeginDirectRect
================
*/
void D_BeginDirectRect (int x, int y, byte *pbitmap, int width, int height)
{
// direct drawing of the "accessing disk" icon isn't supported under Linux
}

/*
================
D_EndDirectRect
================
*/
void D_EndDirectRect (int x, int y, int width, int height)
{
// direct drawing of the "accessing disk" icon isn't supported under Linux
}


/*
===========
HandleEvents

getting and handle events
===========
*/
void HandleEvents (void)
{
	XEvent x_event;

	qboolean dowarp = false;

	if (!x_disp)
		return;

	while (XPending(x_disp)) 
	{
		XNextEvent(x_disp, &x_event);

		switch (x_event.type) 
		{
			case KeyPress: // key pressed
			case KeyRelease: // key released
				Key_Event(XLateKey(&x_event.xkey), x_event.type == KeyPress);
				break;

			case MotionNotify: // mouse moved
				if (mouse_grab_active) 
				{
					if (dga_mouse_active) 
					{
						mouse_x += x_event.xmotion.x_root;
						mouse_y += x_event.xmotion.y_root;
					} 
					else 
					{
						if (!x_event.xmotion.send_event)
						{
							mouse_x += x_event.xmotion.x - p_mouse_x;
							mouse_y += x_event.xmotion.y - p_mouse_y;
							if (abs(vid.width/2 - x_event.xmotion.x) > vid.width / 4 || abs(vid.height/2 - x_event.xmotion.y) > vid.height / 4)
								dowarp = true;
						}
						p_mouse_x = x_event.xmotion.x;
						p_mouse_y = x_event.xmotion.y;
					}
				}
				break;

			case ButtonPress: // mouse button pressed
			case ButtonRelease: // mouse button released
				switch (x_event.xbutton.button)
				{
					case 1:
						Key_Event (K_MOUSE1, x_event.type == ButtonPress);
						break;

					case 2:
						Key_Event (K_MOUSE3, x_event.type == ButtonPress);
						break;

					case 3:
						Key_Event (K_MOUSE2, x_event.type == ButtonPress);
						break;

					case 4:
						Key_Event (K_MWHEELUP, x_event.type == ButtonPress);
						break;

					case 5:
						Key_Event (K_MWHEELDOWN, x_event.type == ButtonPress);
						break;

					case 6:
						Key_Event (K_MOUSE4, x_event.type == ButtonPress);
						break;

					case 7:
						Key_Event (K_MOUSE5, x_event.type == ButtonPress);
						break;

					case 8:
						Key_Event (K_MOUSE6, x_event.type == ButtonPress);
						break;

					case 9:
						Key_Event (K_MOUSE7, x_event.type == ButtonPress);
						break;

					case 10:
						Key_Event (K_MOUSE8, x_event.type == ButtonPress);
						break;
				}
				break;

			case CreateNotify: // window created
				win_x = x_event.xcreatewindow.x;
				win_y = x_event.xcreatewindow.y;
				break;

			case ConfigureNotify: // window changed size/location
				win_x = x_event.xconfigure.x;
				win_y = x_event.xconfigure.y;
				win_width = x_event.xconfigure.width;
				win_height = x_event.xconfigure.height;
				win_notify = true;
				break;

			case DestroyNotify: // window has been destroyed
				Sys_Quit();
				break; 

			case ClientMessage: // window manager messages
				if ((x_event.xclient.format == 32) && ((unsigned int)x_event.xclient.data.l[0] == wm_delete_window_atom))
					Sys_Quit();
				break; 

			case MapNotify: // window restored
				vid_hidden = false;
				if(vidmode_fullscreen)
				{
					// set our video mode
					XF86VidModeSwitchToMode(x_disp, scrnum, &game_vidmode);

					// Move the viewport to top left
					XF86VidModeSetViewPort(x_disp, scrnum, 0, 0);
				}
				break;

			case UnmapNotify: // window iconified/rolledup/whatever
				vid_hidden = true;
				if(vidmode_fullscreen)
				{
					// set our video mode
					XF86VidModeSwitchToMode(x_disp, scrnum, &init_vidmode);
				}
				break;

			case FocusIn: // window is now the input focus
				vid_activewindow = true;
				break; 

			case FocusOut: // window is no longer the input focus
				vid_activewindow = false;
				break; 

			case EnterNotify: // mouse entered window
				vid_notifywindow = true;
				break;

			case LeaveNotify: // mouse left window
				vid_notifywindow = false;
				break;
		}
	}

	if (dowarp)
	{
		/* move the mouse to the window center again */
		p_mouse_x = vid.width / 2;
		p_mouse_y = vid.height / 2;
		XWarpPointer(x_disp, None, x_win, 0, 0, 0, 0, p_mouse_x, p_mouse_y);
	}
}

/*
===========
Sys_SendKeyEvents
===========
*/
void Sys_SendKeyEvents (void)
{
	static qboolean active = true;

	// do some checks, if windowed
	if (!vidmode_fullscreen)
	{
		// check for resize
		if (win_notify)
		{
			win_notify = false;

			if (win_width < 320)
				win_width = 320;
			if (win_height < 200)
				win_height = 200;

			vid.width = win_width;
			vid.height = win_height;

			vid.conwidth = vid.width;
			vid.conheight = vid.height;
			vid.aspect = ((float)vid.height / (float)vid.width) * (320.0 / 240.0);

			vid.recalc_refdef = true; // force a surface cache flush
		}

		// enable/disable sound, set/restore gamma and grab/ungrab keyb
		// on focus gain/loss
		if (!vid_hidden && vid_activewindow)
		{
			if (!active)
			{
				S_UnblockSound ();
				S_ClearBuffer ();
				VID_Gamma_Set ();
				IN_GrabKeyboard();
				active = true;
			}
		}
		else
		{
			if (active)
			{
				S_BlockSound ();
				S_ClearBuffer ();
				VID_Gamma_Restore ();
				IN_UngrabKeyboard();
				active = false;
			}
		}

		// handle the mouse state when windowed if that's changed
		if (key_dest != key_console /* && !mouse_grab_active */ && vid_activewindow)
		{
			IN_GrabMouse ();
		}
		else if ( /* mouse_grab_active && */ key_dest == key_console) 
		{
			IN_UngrabMouse ();
		}
	}

	HandleEvents ();
}

//====================================

/*
=================
GL_BeginRendering
=================
*/
void GL_BeginRendering (int *x, int *y, int *width, int *height)
{
	*x = *y = 0;

	*width = vid.width;
	*height = vid.height;
}

/*
=================
GL_EndRendering
=================
*/
void GL_EndRendering (void)
{
	glFlush();
	glXSwapBuffers(x_disp, x_win);

	if (fullsbardraw)
		Sbar_Changed();
}

/*
===============
VID_SetPalette
===============
*/
void VID_SetPalette (byte *palette)
{
	GL_SetPalette (palette);
}

//====================================

/*
===============
VID_ShiftPalette
===============
*/
void VID_ShiftPalette (byte *p)
{
//	VID_SetPalette(p);
}

/*
===============
VID_Init
===============
*/
void VID_Init(byte *palette)
{
	int i;
	int attrib[] = {
		GLX_RGBA,
		GLX_RED_SIZE, 1,
		GLX_GREEN_SIZE, 1,
		GLX_BLUE_SIZE, 1,
		GLX_DOUBLEBUFFER,
		GLX_DEPTH_SIZE, 1,
		None
	};

	XSetWindowAttributes attr;
	XClassHint *clshints;
	XWMHints *wmhints;
	XSizeHints *szhints;
	unsigned long mask;
	Window root;
	XVisualInfo *visinfo;
	qboolean fullscreen = true;
	qboolean vidmode_ext = false;
	qboolean VidMode = false;
	int MajorVersion, MinorVersion;

	Cvar_RegisterVariable (&vid_mode);

// set vid parameters
	vid.width = 640;
	vid.height = 480;
	vid.maxwarpwidth = WARP_WIDTH;
	vid.maxwarpheight = WARP_HEIGHT;
	vid.numpages = 2;
	vid.colormap = host_colormap;
	vid.fullbright = 256 - LittleLong (*((int *)vid.colormap + 2048));

// open the display
	x_disp = XOpenDisplay(NULL);
	if (!x_disp) 
	{
		if (getenv("DISPLAY"))
			Sys_Error("VID_Init: Could not open display [%s]\n", getenv("DISPLAY"));
		else
			Sys_Error("VID_Init: Could not open local display\n");
	}

// get video mode list
	MajorVersion = MinorVersion = 0;
	VidMode = XF86VidModeQueryVersion(x_disp, &MajorVersion, &MinorVersion);
//	if (XF86VidModeQueryVersion(x_disp, &MajorVersion, &MinorVersion)) 
	if (COM_CheckParm("-novidmode"))
	{
		Con_Warning ("XFree86 VidMode extension disabled at command line\n");
	}
	else if (VidMode)
	{
		Con_Printf("XFree86 VidMode extension version %d.%d found\n", MajorVersion, MinorVersion);
		vidmode_ext = true;
	}
	else
	{
		Con_Warning ("XFree86 VidMode extension not supported\n");
	}

// get visual
	scrnum = DefaultScreen(x_disp);
	root = RootWindow(x_disp, scrnum);
	visinfo = glXChooseVisual(x_disp, scrnum, attrib);
	if (!visinfo) 
	{
		Sys_Error ("VID_Init: couldn't get an RGB, Double-buffered, Depth visual\n");
	}

// check for command-line size parameters
	if ((i = COM_CheckParm("-width"))) 
	{
		if (i >= com_argc-1)
			Sys_Error("VID_Init: -width <width>\n");

		vid.width = atoi(com_argv[i+1]);
		if (!vid.width)
			Sys_Error("VID_Init: Bad width\n");
	}
	if ((i = COM_CheckParm("-height"))) 
	{
		if (i >= com_argc-1)
			Sys_Error("VID_Init: -height <height>\n");

		vid.height = atoi(com_argv[i+1]);
		if (!vid.height)
			Sys_Error("VID_Init: Bad height\n");
	}

// check for command-line video parameters
	if (COM_CheckParm("-current"))
	{
		vid.width = DisplayWidth (x_disp, scrnum);
		vid.height = DisplayHeight (x_disp, scrnum);
	}
	else if (COM_CheckParm("-window"))
	{
		fullscreen = false;
	}

// set the vidmode to the requested width/height if possible 
	if (vidmode_ext) 
	{
		int x, y, dist;
		int best_dist = 9999999;
		int num_modes;
		XF86VidModeModeInfo **modes;
		XF86VidModeModeInfo *mode = NULL;

		XF86VidModeGetAllModeLines(x_disp, scrnum, &num_modes, &modes);

		// are we going fullscreen?  If so, let's change video mode
		if (fullscreen) 
		{
			for (i = 0; i < num_modes; i++) 
			{
				if (vid.width > modes[i]->hdisplay || vid.height > modes[i]->vdisplay)
					continue;

				x = vid.width - modes[i]->hdisplay;
				y = vid.height - modes[i]->vdisplay;
				dist = (x * x) + (y * y);

				if (dist < best_dist) 
				{
					best_dist = dist;
					mode = modes[i];
				}
			}

			if (mode)
			{
				// change to the mode
				XF86VidModeSwitchToMode(x_disp, scrnum, mode);
				memcpy(&game_vidmode, mode, sizeof(game_vidmode)); // save game vidmode
				memcpy(&init_vidmode, modes[0], sizeof(init_vidmode)); // save system initial vidmode
				vidmode_fullscreen = true;

				// Move the viewport to top left
				XF86VidModeSetViewPort(x_disp, scrnum, 0, 0);
				Con_DPrintf ("VID_Init: set fullscreen mode at %dx%d\n", vid.width, vid.height); 
			} 
			else
			{
//				fullscreen = false; // eh?
				Con_DPrintf ("VID_Init: unable to set fullscreen mode at %dx%d\n", vid.width, vid.height); 
			}

			free(modes);
		}
	}

// window attributes
	mask = CWBackPixel | CWColormap | CWEventMask;
	attr.background_pixel = 0;
	attr.colormap = XCreateColormap(x_disp, root, visinfo->visual, AllocNone);
	attr.event_mask = X_MASK;

// setup attributes for main window
	if (vidmode_fullscreen) 
	{
		mask |= CWSaveUnder | CWBackingStore | CWOverrideRedirect;
		attr.override_redirect = True;
		attr.backing_store = NotUseful;
		attr.save_under = False;
	} 
	else
	{
		mask |= CWBorderPixel;
		attr.border_pixel = 0;
	}

// create the main window
	x_win = XCreateWindow (
		x_disp, 
		root, 
		0, 0, // x, y
		vid.width, vid.height, 
		0, // borderwidth
		visinfo->depth, 
		InputOutput, 
		visinfo->visual, 
		mask, 
		&attr
	);

	wmhints = XAllocWMHints();
	if(XpmCreatePixmapFromData(x_disp, x_win, quake_xpm, &wmhints->icon_pixmap, &wmhints->icon_mask, NULL) == XpmSuccess)
		wmhints->flags |= IconPixmapHint | IconMaskHint;

	clshints = XAllocClassHint();
	clshints->res_name = strdup("SourceQuake");
	clshints->res_class = strdup("SourceQuake");

	szhints = XAllocSizeHints();
	if(vidmode_fullscreen)
	{
		szhints->min_width = szhints->max_width = vid.width;
		szhints->min_height = szhints->max_height = vid.height;
		szhints->flags |= PMinSize | PMaxSize;
	}

	XmbSetWMProperties(x_disp, x_win, "SourceQuake", "SourceQuake", (char **) com_argv, com_argc, szhints, wmhints, clshints);
	XFree(clshints);
	XFree(wmhints);
	XFree(szhints);

//	if (!vidmode_fullscreen)
//		XStoreName(x_disp, x_win, "SourceQuake");

// map the window
	XMapWindow(x_disp, x_win);

// making the close button on a window do the right thing
// seems to involve this mess, sigh...
	wm_delete_window_atom = XInternAtom(x_disp, "WM_DELETE_WINDOW", false);
	XSetWMProtocols(x_disp, x_win, &wm_delete_window_atom, 1);

	if (vidmode_fullscreen) 
	{
		XMoveWindow(x_disp, x_win, 0, 0);
		XRaiseWindow(x_disp, x_win);
		XWarpPointer(x_disp, None, x_win, 0, 0, 0, 0, 0, 0);
		XFlush(x_disp);

		// Move the viewport to top left
		XF86VidModeSetViewPort(x_disp, scrnum, 0, 0);
	}

//	XFlush(x_disp); // disabled because XSync later

	ctx = glXCreateContext(x_disp, visinfo, NULL, True);
	glXMakeCurrent(x_disp, x_win, ctx);

	XSync(x_disp, False); // from darkplaces test it

	vid_activewindow = true;
	vid_hidden = false;
	vid_notifywindow = true;

	vid.conwidth = vid.width;
	vid.conheight = vid.height;
	vid.aspect = ((float)vid.height / (float)vid.width) * (320.0 / 240.0);

	InitSig(); // trap signals
	GL_Init();
	VID_SetPalette(palette);
	VID_Gamma_Init ();

	vid.recalc_refdef = true; // force a surface cache flush

	if (COM_CheckParm("-fullsbar"))
		fullsbardraw = true;

	Con_SafePrintf ("Video mode %dx%d %s initialized\n", vid.width, vid.height, vidmode_fullscreen ? "fullscreen" : "windowed");
}

/*
===============
VID_Shutdown

called at shutdown
===============
*/
void VID_Shutdown (void)
{
	if (x_disp) 
	{
		if (ctx)
			glXDestroyContext(x_disp, ctx);

		VID_Gamma_Shutdown ();

		if (x_win)
			XDestroyWindow(x_disp, x_win);

		if (vidmode_fullscreen)
			XF86VidModeSwitchToMode(x_disp, scrnum, &init_vidmode);

		XCloseDisplay (x_disp);
	}

	vidmode_fullscreen = false;
	x_disp = NULL;
	x_win = 0;
	ctx = NULL;
}

/*
===============
VID_LockBuffer
===============
*/
void VID_LockBuffer (void)
{
}

/*
===============
VID_UnlockBuffer
===============
*/
void VID_UnlockBuffer (void)
{
}

